use std::collections::HashMap;
use std::fs::OpenOptions;
use std::io::BufWriter;
use std::io::Write;
use strfmt::strfmt;

use super::*;
use crate::parse_deser::*;
use crate::Args;

fn gen_ninja_header(config: &Config, cmdline_args: &Args) -> Result<String> {
    let mut ret = "".to_string();
    ret.push_str(&format!(
        "## Generated by ngen
## Do not modify by hand

builddir = {0}

rule mkdir
  command = mkdir -p $out
  description = Creating directory $out

rule regen_ninjafile
  command = ngen -c $in -o $out
  generator = 1
  description = Regenerating $out

",
        config.build_dir
    ));
    if config.compile_commands {
        ret.push_str(&format!(
            "\
rule regen_compile_commands
  command = ngen -c $in --write-compile-commands $out
  description = Regenerating $out

build $builddir: mkdir

build $builddir/compile_commands.json: regen_compile_commands {1} || $builddir
  pool = console

build {0}: regen_ninjafile {1} || $builddir/compile_commands.json
  pool = console
",
            cmdline_args.output_file, cmdline_args.config_file
        ));
    } else {
        ret.push_str(&format!(
            "build {}: regen_ninjafile {}\n  pool = console\n",
            cmdline_args.output_file, cmdline_args.config_file
        ));
    }
    Ok(ret)
}

fn gen_ninja_target(name: &str, target: &Target) -> Result<String> {
    if target.sources.len() == 0 {
        return Err(Error {
            kind: ErrorKind::NoSources,
            message: format!("target `{}` has no sources", name),
        });
    }

    let mut ret = "\n".to_string();

    let compiler_flags = target.compiler_flags.join(" ");
    let linker_flags = target.linker_flags.join(" ");
    let linker_libs = target.linker_libs.join(" ");

    let mut fmt_args = HashMap::new();
    fmt_args.insert("outfile".to_string(), "$out");
    fmt_args.insert("compiler".to_string(), target.compiler);
    fmt_args.insert("compiler_flags".to_string(), &compiler_flags);
    fmt_args.insert("linker".to_string(), target.compiler);
    fmt_args.insert("object".to_string(), "$out");
    fmt_args.insert("objects".to_string(), "$in");
    fmt_args.insert("source".to_string(), "$in");
    fmt_args.insert("depfile".to_string(), "$dep");
    fmt_args.insert("linker_flags".to_string(), &linker_flags);
    fmt_args.insert("linker_libs".to_string(), &linker_libs);

    ret.push_str(&format!(
        "\
# BEGIN TARGET {name}
rule cc_{name}
  deps = gcc
  depfile = $dep
  command = "
    ));
    match strfmt(target.opts.compile_cmd_fmt, &fmt_args) {
        Ok(v) => ret.push_str(&v),
        Err(e) => {
            return Err(Error {
                kind: ErrorKind::FormatError,
                message: e.to_string(),
            });
        }
    }
    ret.push_str(&format!(
        "
  description = Building $in -> $out
rule link_{name}
  command = "
    ));
    match strfmt(target.opts.link_cmd_fmt, &fmt_args) {
        Ok(v) => ret.push_str(&v),
        Err(e) => {
            return Err(Error {
                kind: ErrorKind::FormatError,
                message: e.to_string(),
            });
        }
    }
    ret.push_str(&format!(
        "
  description = Linking $out

build $builddir/{name}/obj: mkdir
build $builddir/{name}/dep: mkdir

"
    ));
    let mut object_list: Vec<String> = Vec::new();

    for source in &target.sources {
        let new_file = source.replace("/", "-");
        let obj_name = format!("$builddir/{}/obj/{}.o", name, new_file);
        let dep_name = format!("$builddir/{}/dep/{}.o.d", name, new_file);
        ret.push_str(&format!(
            "build {}: cc_{} {}\n  dep = {}\n",
            obj_name, name, source, dep_name
        ));
        object_list.push(obj_name);
    }

    ret.push_str(&format!(
        "\nbuild $builddir/{0}/{1}: link_{0} ",
        name, target.outfile
    ));
    ret.push_str(&object_list.join(" "));
    let deps = &target.opts.depend_on.join(" ");
    ret.push_str(&format!(
        " | {2} || $builddir/{0}/obj $builddir/{0}/dep\nbuild {0}: phony $builddir/{0}/{1}\n",
        name, target.outfile, deps
    ));
    if target.opts.default {
        ret.push_str(&format!("\ndefault {}\n", name));
    }
    ret.push_str(&format!("# END TARGET {}\n", name));
    Ok(ret)
}

pub fn create_ninja_file_content(
    target_list: &TargetList,
    config: &Config,
    args: &Args,
) -> Result<String> {
    let mut ret = String::new();
    ret.push_str(&gen_ninja_header(&config, &args)?);
    for (name, target) in &target_list.0 {
        let content = gen_ninja_target(name, target);
        match content {
            Ok(v) => ret.push_str(&v),
            Err(e) => match e.kind {
                ErrorKind::NoSources => eprintln!("WARNING: {e}. Skipping `{}`", name),
                _ => {
                    return Err(e);
                }
            },
        }
    }

    Ok(ret)
}

pub fn write_ninja_file(filename: &str, content: &[u8]) -> std::io::Result<()> {
    let file = OpenOptions::new()
        .write(true)
        .truncate(true)
        .create(true)
        .open(filename)?;

    let mut writer = BufWriter::new(file);
    writer.write(content)?;
    Ok(())
}
