use std::{collections::HashMap, fmt::Display};

use crate::parse_deser::*;

pub enum ErrorKind {
    NoSources,
    DependencyDoesNotExist,
    DependencyCycle,
    ParentDoesNotExist,
    InheritanceCycle,
    CompileCmdsTargetDoesNotExist,
}

pub struct Error {
    pub kind: ErrorKind,
    pub message: String,
}

impl Display for Error {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "configuration error: {}", self.message)
    }
}

type Result<T> = std::result::Result<T, Error>;

type Chain<'a> = HashMap<&'a str, &'a str>;

fn print_chain(chain: &Chain, start: &str, connector: &str) -> String {
    let mut ret = "target".to_string();
    let mut i = 0;
    let mut start = start;
    while i < chain.len() {
        let dep = chain.get(start).unwrap();
        if i == chain.len() - 1 {
            ret.push_str(&format!(" but target `{}` {} `{}`", start, connector, dep));
            break;
        }
        ret.push_str(&format!(" `{}` {} `{}`,", start, connector, dep));
        start = dep;
        i += 1;
    }

    ret
}

impl<'a> TargetList<'a> {
    pub fn check_if_any_sources_are_provided(&self) -> Result<()> {
        let mut sources_found = false;
        for target in self.0.values() {
            if !target.sources.is_empty() {
                sources_found = true;
                break;
            }
        }
        if sources_found {
            Ok(())
        } else {
            Err(Error {
                kind: ErrorKind::NoSources,
                message: "no targets have any sources".to_string(),
            })
        }
    }
}

impl<'a> TargetOptionList<'a> {
    fn check_target_dep(
        &self,
        opts: &TargetOptions<'a>,
        name: &'a str,
        dep_chain: &mut Chain<'a>,
    ) -> Result<()> {
        for dep in opts.depend_on.clone() {
            match self.0.get(dep) {
                Some(v) => {
                    dep_chain.insert(name, dep);
                    if let Some(_) = dep_chain.get(dep) {
                        return Err(Error {
                            kind: ErrorKind::DependencyCycle,
                            message: format!(
                                "dependency cycle detected: {}",
                                print_chain(dep_chain, name, "depends on")
                            ),
                        });
                    }
                    self.check_target_dep(v, dep, dep_chain)?;
                    let _ = dep_chain.remove(name);
                }
                None => {
                    return Err(Error {
                        kind: ErrorKind::DependencyDoesNotExist,
                        message: format!(
                            "target `{0}` depends on `{1}`, but target `{1}` does not exist",
                            name, dep
                        ),
                    });
                }
            }
        }
        Ok(())
    }

    pub fn check_for_dependency_cycle(&self) -> Result<()> {
        for (name, opts) in &self.0 {
            let mut dep_chain: Chain = HashMap::new();
            self.check_target_dep(opts, name, &mut dep_chain)?;
        }
        Ok(())
    }

    fn check_target_parent(
        &self,
        opts: &TargetOptions<'a>,
        name: &'a str,
        dep_chain: &mut Chain<'a>,
    ) -> Result<()> {
        let inherit = opts.inherit_from;
        if name == inherit {
            return Ok(());
        }
        match self.0.get(inherit) {
            Some(v) => {
                dep_chain.insert(name, inherit);
                if let Some(_) = dep_chain.get(inherit) {
                    return Err(Error {
                        kind: ErrorKind::InheritanceCycle,
                        message: format!(
                            "inheritance cycle detected: {}",
                            print_chain(dep_chain, name, "inherits from")
                        ),
                    });
                }
                self.check_target_parent(v, inherit, dep_chain)?;
            }
            None => {
                return Err(Error {
                    kind: ErrorKind::ParentDoesNotExist,
                    message: format!(
                        "target `{0}` inherits from `{1}`, but target `{1}` does not exist",
                        name, inherit
                    ),
                });
            }
        }
        Ok(())
    }

    pub fn get_inheritance_order(&self) -> Result<Vec<Vec<String>>> {
        let mut ret: Vec<Vec<String>> = vec![];
        for (name, opts) in &self.0 {
            if !opts.inherit {
                ret.push(vec![name.to_string()]);
                continue;
            }
            let mut dep_chain: Chain = HashMap::new();

            self.check_target_parent(opts, name, &mut dep_chain)?;

            let mut order: Vec<String> = vec![name.to_string()];
            let mut i = 0;
            while let Some(v) = dep_chain.get(&order[i] as &str) {
                order.push(v.to_string());
                i += 1;
            }
            ret.push(order);
        }
        Ok(ret)
    }
}

impl<'a> Config<'a> {
    pub fn check_compile_commands_target(&self, target_list: &TargetList) -> Result<()> {
        if !self.compile_commands {
            return Ok(());
        }
        match target_list.0.get(self.compile_commands_target) {
            Some(_) => Ok(()),
            None => Err(Error {
                kind: ErrorKind::CompileCmdsTargetDoesNotExist,
                message: format!(
                    "compile commands target `{}` does not exist",
                    self.compile_commands_target
                ),
            }),
        }
    }
}
